# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging
import os
import re
import shutil
from subprocess import Popen, PIPE, TimeoutExpired


try:
    from ansible_builder import introspect
except ImportError:
    from ansible_builder._target_scripts import introspect

from galaxy_importer import exceptions as exc
from galaxy_importer.finder import ContentFinder, FileWalker
from galaxy_importer import constants
from galaxy_importer import loaders, file_parser, schema
from galaxy_importer.utils import markup as markup_utils
from galaxy_importer.utils import chksums
from galaxy_importer.utils import string_utils

default_logger = logging.getLogger(__name__)

DOCUMENTATION_DIR = "docs"


class CollectionLoader(object):
    """Loads collection and content info."""

    def __init__(self, path, filename, cfg=None, logger=None):
        self.log = logger or default_logger
        self.path = path
        self.filename = filename
        self.cfg = cfg

        self.content_objs = None
        self.metadata = None
        self.file_manifest_file = None
        self.docs_blob = None
        self.contents = None

        # build the collections path for lint's module resolution
        self.collections_path = self.path
        if not callable(self.path):
            if hasattr(self.path, "strpath"):
                paths = self.path.strpath.split(os.sep)
            else:
                paths = self.path.split(os.sep)
            if "ansible_collections" in paths:
                ix = paths.index("ansible_collections")
                self.collections_path = os.sep.join(paths[: ix + 1])

    def load(self):
        # NOTE: If we knew the chksum for MANIFEST.json, we could check it here first
        self.manifest = self._load_manifest()

        self.metadata = self.manifest.collection_info

        # The default name for 'file_manifest_file' is FILES.json
        self.file_manifest_file = self.manifest.file_manifest_file

        # load data from FILES.json
        self.file_manifest = self._load_file_manifest(
            path_prefix=self.path, file_manifest_file=self.file_manifest_file
        )

        # check chksum for each file in FILES.json
        # Note: Will raise exceptions on file_manifest / FILES.json errors
        self._check_file_manifest(self.path, self.file_manifest, self.file_manifest_file.name)

        self._rename_extract_path()
        self._check_filename_matches_manifest()
        self._check_metadata_filepaths()

        self.doc_strings = {}
        if self.cfg.run_ansible_doc:
            self.doc_strings = loaders.DocStringLoader(
                path=self.path,
                fq_collection_name="{}.{}".format(self.metadata.namespace, self.metadata.name),
                logger=self.log,
                cfg=self.cfg,
            ).load()

        self.content_objs = list(self._load_contents())

        self.contents = self._build_contents_blob()
        self.docs_blob = self._build_docs_blob()
        self.requires_ansible = file_parser.RuntimeFileParser(self.path).get_requires_ansible()
        self._check_ee_yml_dep_files()
        self._check_collection_changelog()

        if self.cfg.run_ansible_lint:
            self._lint_collection()
        self._check_ansible_test_ignore_files()

        return schema.ImportResult(
            metadata=self.metadata,
            docs_blob=self.docs_blob,
            contents=self.contents,
            requires_ansible=self.requires_ansible,
        )

    def _lint_collection(self):
        """Log ansible-lint output.

        ansible-lint stdout are linter violations, they are logged as warnings and errors,
        depending on the rule level.

        ansible-lint stderr includes info about vars, file discovery,
        summary of linter violations, config suggestions, and raised errors.
        Only raised errors are logged, they are logged as errors.
        """

        self.log.info("Linting collection via ansible-lint...")

        if not shutil.which("ansible-lint"):
            self.log.warning("ansible-lint not found, skipping lint of collection")
            return

        cmd = [
            "/usr/bin/env",
            f"ANSIBLE_COLLECTIONS_PATH={self.collections_path}",
            f"ANSIBLE_LOCAL_TEMP={self.cfg.ansible_local_tmp}",
            "ansible-lint",
            "--profile",
            "production",
            "--exclude",
            "tests/integration/",
            "--exclude",
            "tests/unit/",
            "--parseable",
            "--nocolor",
        ]
        if self.cfg.offline_ansible_lint:
            cmd.append("--offline")

        self.log.debug("CMD: " + " ".join(cmd))
        proc = Popen(
            cmd,
            cwd=self.path,
            encoding="utf-8",
            stdout=PIPE,
            stderr=PIPE,
        )

        try:
            outs, errs = proc.communicate(timeout=180)
        except (
            TimeoutExpired
        ):  # pragma: no cover - a TimeoutExpired mock would apply to both calls to commnicate()
            self.log.error("Timeout on call to ansible-lint")
            proc.kill()
            outs, errs = proc.communicate()

        for line in outs.splitlines():
            self.log.warning(line.strip())

        for line in errs.splitlines():
            if line.startswith(constants.ANSIBLE_LINT_ERROR_PREFIXES):
                self.log.warning(line.rstrip())

        # The prevous code tries to be intelligent about what to display or not display
        # but we have serious errors from lint that are hidden by that logic. We should
        # attempt to inform the users of these errors (especially tracebacks).
        if proc.returncode != 0 and errs and "Traceback (most recent call last):" in errs:
            self.log.error(errs)

        self.log.info("...ansible-lint run complete")

    def _check_ansible_test_ignore_files(self):  # pragma: no cover
        """Log a warning when ansible test sanity ignore files are present.
        Method excluded from pytest coverage, test exist outside repo via iqe.
        """
        IGNORE_FILE_REGEXP = re.compile(r"^ignore-.+\.txt$")
        IGNORE_WARNING = (
            "Ignore files skip ansible-test sanity tests, "
            "found {file} with {line_count} statement(s)"
        )

        sanity_path = os.path.join(self.path, "tests", "sanity")
        if not os.path.exists(sanity_path):
            return

        listdir = os.listdir(sanity_path)
        for ignore_file in filter(IGNORE_FILE_REGEXP.match, listdir):
            with open(os.path.join(sanity_path, ignore_file), "r") as f:
                line_count = len(f.readlines())
            self.log.warning(IGNORE_WARNING.format(file=ignore_file, line_count=line_count))

    def _check_collection_changelog(self):
        """Log an error when a CHANGELOG file is not present in the root of the collection."""
        changelog_rst_path = os.path.join(self.path, "CHANGELOG.rst")
        changelog_md_path = os.path.join(self.path, "CHANGELOG.md")
        changelog_yaml_path = os.path.join(self.path, "changelogs/changelog.yaml")

        if (
            not os.path.exists(changelog_rst_path)
            and not os.path.exists(changelog_md_path)
            and not os.path.exists(changelog_yaml_path)
        ):
            self.log.warning(
                "No changelog found. "
                "Add a CHANGELOG.rst, CHANGELOG.md, or changelogs/changelog.yaml file."
            )

    def _check_ee_yml_dep_files(self):  # pragma: no cover
        """Check for python deps file and system deps file if they are listed in
        meta/execution-environment.yml, and log if listed files are not found

        Method excluded from pytest coverage, test exist outside repo via iqe.
        """

        try:
            python_deps, system_deps = introspect.process_collection(self.path)
        except FileNotFoundError as e:
            self.log.warning(
                f"Error when checking meta/execution-environment.yml for dependency files: {e}"
            )

    def _load_manifest(self):
        manifest_file = os.path.join(self.path, "MANIFEST.json")
        if not os.path.exists(manifest_file):
            raise exc.ManifestNotFound("No manifest found in collection")

        default_logger.debug("manifest_file: %s", manifest_file)

        with open(manifest_file, "r") as f:
            try:
                data = schema.CollectionArtifactManifest.parse(f.read())
            except ValueError as e:
                raise exc.ManifestValidationError(str(e)) from e

            default_logger.debug("data: %s", data)
            default_logger.debug("data.file_manifest_file: %s", data.file_manifest_file)
            return data

    def _load_file_manifest(self, path_prefix, file_manifest_file):
        """Load CollectionArtifactFileManifest data from file_manifest_file

        Args:
            path_prefix (str): Any file path prefix we need to add to file paths in the
                CollectionArtifactFile artifact_file
            file_manifest_file (CollectionArtifactFile): object with info about the
                FILES.json in the artifact. The info includes the name, type, path,
                and chksum.

        Raises:
            ManifestValidationError: If there are any errors loading, parsing,
                deserializing, or validating the data in file_manifest_file

        Returns:
            CollectionArtifactFileManifest: The data from file_manifest_file

        """
        default_logger.debug("file_manifest_file: %s", file_manifest_file)

        chksums.check_artifact_file(path_prefix=path_prefix, artifact_file=file_manifest_file)

        files_manifest_file = os.path.join(path_prefix, file_manifest_file.name)
        default_logger.debug("files_manifest_file: %s", files_manifest_file)

        with open(files_manifest_file, "r") as f:
            try:
                file_manifest = schema.CollectionArtifactFileManifest.parse(f.read())
            except ValueError as e:
                raise exc.ManifestValidationError(str(e))

        return file_manifest

    def _check_file_manifest(self, path_prefix, file_manifest, file_manifest_name):
        """Check the file content described in file_manifest

        Check the chksums for files.
        Check for any missing files.
        Check for any missing dirs.

        Args:
            path_prefix (str): Any file path prefix we need to add to file paths in the
                CollectionArtifactFile artifact_file
            file_manifest (CollectionArtifactFileManifest): Object with list of
                CollectionArtifactFile items. The CollectionArtifactFile data
                is used to check and validate each file.
            file_manifest_name (str): The name of the file manifest (ie, "FILES.json")

        Raises:
            CollectionArtifactFileNotFound: If artifact_file is not found on the file system.
            CollectionArtifactFileChecksumError: If the sha256sum of the on disk
                artifact_file contents does not match artifact_file.chksum_sha256.

        Returns:
            bool: All the items in file_manifest were found and valid
        """

        for artifact_file in file_manifest.files:
            if artifact_file.ftype != "file":
                continue

            chksums.check_artifact_file(path_prefix=path_prefix, artifact_file=artifact_file)

        # check the extract archive for any extra files.
        filewalker = FileWalker(collection_path=path_prefix)
        prefix = path_prefix + "/"
        found_file_set = set([string_utils.removeprefix(fp, prefix) for fp in filewalker.walk()])

        file_manifest_file_set = set([artifact_file.name for artifact_file in file_manifest.files])
        # The artifact contains MANIFEST.json and FILES.JSON, but they aren't
        # in file list in FILES.json so add them so we match expected.
        file_manifest_file_set.add("MANIFEST.json")
        file_manifest_file_set.add(file_manifest_name)

        difference = sorted(list(found_file_set.difference(file_manifest_file_set)))

        if difference:
            err_msg = f"Files in the artifact but not the file manifest: {difference}"
            raise exc.FileNotInFileManifestError(unexpected_files=difference, msg=err_msg)

        return True

    def _rename_extract_path(self):
        old_ns_dir = os.path.dirname(self.path)
        ansible_collections_dir = os.path.dirname(old_ns_dir)
        new_ns_dir = os.path.join(ansible_collections_dir, self.metadata.namespace)
        os.rename(old_ns_dir, new_ns_dir)

        old_name_dir = os.path.join(new_ns_dir, os.path.basename(self.path))
        new_name_dir = os.path.join(new_ns_dir, self.metadata.name)
        os.rename(old_name_dir, new_name_dir)
        self.path = new_name_dir
        self.log.debug(f"Renamed extract dir to: {self.path}")

    def _check_filename_matches_manifest(self):
        if not self.filename:
            return
        for item in ["namespace", "name", "version"]:
            filename_item = getattr(self.filename, item, None)
            metadata_item = getattr(self.metadata, item, None)
            if not filename_item:
                continue
            if filename_item != metadata_item:
                raise exc.ManifestValidationError(
                    f'Filename {item} "{filename_item}" did not match metadata "{metadata_item}"'
                )

    def _load_contents(self):
        """Find and load data for each content inside the collection."""
        found_contents = ContentFinder().find_contents(self.path, self.log)
        for content_type, rel_path in found_contents:
            loader_cls = loaders.get_loader_cls(content_type)
            loader = loader_cls(
                content_type, rel_path, self.path, self.doc_strings, self.cfg, self.log
            )
            content_obj = loader.load()

            yield content_obj

    def _build_contents_blob(self):
        """Build importer result contents from Content objects."""
        return [
            schema.ResultContentItem(
                name=c.name,
                content_type=c.content_type.value,
                description=c.description,
            )
            for c in self.content_objs
        ]

    def _build_docs_blob(self):
        """Build importer result docs_blob from collection documentation."""

        # return an empty DocsBlob if run_ansible_doc=False
        rendered_readme = schema.RenderedDocFile()
        docs_blob = schema.DocsBlob(
            collection_readme=rendered_readme,
            documentation_files=[],
            contents=[],
        )

        if not self.cfg.run_ansible_doc:
            return docs_blob

        contents = [
            schema.DocsBlobContentItem(
                content_name=c.name,
                content_type=c.content_type.value,
                doc_strings=c.doc_strings,
                readme_file=c.readme_file,
                readme_html=c.readme_html,
            )
            for c in self.content_objs
        ]

        readme = markup_utils.get_readme_doc_file(self.path)
        if not readme:
            raise exc.ImporterError("No collection readme found")
        rendered_readme = schema.RenderedDocFile(
            name=readme.name, html=markup_utils.get_html(readme)
        )

        rendered_doc_files = []
        doc_files = markup_utils.get_doc_files(os.path.join(self.path, DOCUMENTATION_DIR))
        if doc_files:
            rendered_doc_files = [
                schema.RenderedDocFile(name=f.name, html=markup_utils.get_html(f))
                for f in doc_files
            ]

        return schema.DocsBlob(
            collection_readme=rendered_readme,
            documentation_files=rendered_doc_files,
            contents=contents,
        )

    def _check_metadata_filepaths(self):
        # NOTE: This may be redundant if _check_file_manifest() looks for missing files
        paths = []
        paths.append(os.path.join(self.path, self.metadata.readme))
        if self.metadata.license_file:
            paths.append(os.path.join(self.path, self.metadata.license_file))
        for path in paths:
            if not os.path.exists(path):
                raise exc.ManifestValidationError(f"Could not find file {os.path.basename(path)}")



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import enum
import re

ANSIBLE_DOC_SUPPORTED_TYPES = [
    "become",
    "cache",
    "callback",
    "cliconf",
    "connection",
    "httpapi",
    "inventory",
    "lookup",
    "shell",
    "module",
    "strategy",
    "vars",
]
ANSIBLE_DOC_PLUGIN_MAP = {"module": "modules"}
ANSIBLE_DOC_KEYS = ["doc", "metadata", "examples", "return"]
ANSIBLE_LINT_ERROR_PREFIXES = ("CRITICAL", "ERROR")
CONTENT_NAME_REGEXP = re.compile(r"^(?!.*__)[a-z_]+[0-9a-z_]*$")
ROLE_META_FILES = ["meta/main.yml", "meta/main.yaml", "meta.yml", "meta.yaml"]
FLAKE8_MAX_LINE_LENGTH = 160
FLAKE8_IGNORE_ERRORS = "E402"
FLAKE8_SELECT_ERRORS = "E,F,W"
MAX_LENGTH_REQUIRES_ANSIBLE = 255
MAX_TAGS_COUNT = 20
NAME_REGEXP = re.compile(r"^(?!.*__)[a-z]+[0-9a-z_]*$")

# For these extensions we support listing them in the galaxy contents list
# In the future we may allow any extension to be listed, and call ansible-doc on it
EDA_EVENT_SOURCE_NAME = "eda/plugins/event_source"
EDA_EVENT_FILTER_NAME = "eda/plugins/event_filter"
ALLOWED_EXTENSION_DIRS = [
    EDA_EVENT_SOURCE_NAME,
    EDA_EVENT_FILTER_NAME,
]

# Match github's allowable usernames, but also include underscores
# because of old-galaxy's preference for replacing hyphens with underscores
# in namespace names. There are quite a few legacy namespaces with
# underscores grandfathered in, so we must support those too.
GITHUB_USERNAME_REGEXP = re.compile(r"^[a-zA-Z\d_](?:[a-zA-Z\d_]|-(?=[a-zA-Z\d])){0,38}$")

# Matches role names with any combination of lowercase letters,
# uppercase letters, numbers, underscores, and hyphens with
# a length in the inclusive range [1, 55].
LEGACY_ROLE_NAME_REGEXP = re.compile("^[a-zA-Z0-9_-]{1,55}$")
# Matches namespaces: Namespace names should match any valid
# GitHub username. Username may only contain alphanumeric characters
# or single hyphens, and cannot begin or end with a hyphen
# Some roles may contain underscores
# Maximum of 39 char tested in the validator
# For retrocompatibility with legacy role, allow names
# finishing with hyphen or underscores, and names containing dots
LEGACY_NAMESPACE_REGEXP = re.compile("^([a-zA-Z0-9.]+[-_]?)+$")


class ContentCategory(enum.Enum):
    MODULE = "module"
    ROLE = "role"
    PLUGIN = "plugin"
    PLAYBOOK = "playbook"
    EXTENSION = "extension"


class ContentType(enum.Enum):
    ROLE = "role"
    MODULE = "module"
    MODULE_UTILS = "module_utils"
    ACTION_PLUGIN = "action"
    BECOME_PLUGIN = "become"
    CACHE_PLUGIN = "cache"
    CALLBACK_PLUGIN = "callback"
    CLICONF_PLUGIN = "cliconf"
    CONNECTION_PLUGIN = "connection"
    DOC_FRAGMENTS_PLUGIN = "doc_fragments"
    FILTER_PLUGIN = "filter"
    HTTPAPI_PLUGIN = "httpapi"
    INVENTORY_PLUGIN = "inventory"
    LOOKUP_PLUGIN = "lookup"
    NETCONF_PLUGIN = "netconf"
    SHELL_PLUGIN = "shell"
    STRATEGY_PLUGIN = "strategy"
    TERMINAL_PLUGIN = "terminal"
    TEST_PLUGIN = "test"
    VARS_PLUGIN = "vars"
    EDA_EVENT_SOURCE = EDA_EVENT_SOURCE_NAME
    EDA_EVENT_FILTER = EDA_EVENT_FILTER_NAME

    @property
    def category(self):
        return {
            ContentType.ROLE: ContentCategory.ROLE,
            ContentType.MODULE: ContentCategory.MODULE,
            ContentType.MODULE_UTILS: ContentCategory.PLUGIN,
            ContentType.ACTION_PLUGIN: ContentCategory.PLUGIN,
            ContentType.BECOME_PLUGIN: ContentCategory.PLUGIN,
            ContentType.CACHE_PLUGIN: ContentCategory.PLUGIN,
            ContentType.CALLBACK_PLUGIN: ContentCategory.PLUGIN,
            ContentType.CLICONF_PLUGIN: ContentCategory.PLUGIN,
            ContentType.CONNECTION_PLUGIN: ContentCategory.PLUGIN,
            ContentType.DOC_FRAGMENTS_PLUGIN: ContentCategory.PLUGIN,
            ContentType.FILTER_PLUGIN: ContentCategory.PLUGIN,
            ContentType.HTTPAPI_PLUGIN: ContentCategory.PLUGIN,
            ContentType.INVENTORY_PLUGIN: ContentCategory.PLUGIN,
            ContentType.LOOKUP_PLUGIN: ContentCategory.PLUGIN,
            ContentType.NETCONF_PLUGIN: ContentCategory.PLUGIN,
            ContentType.SHELL_PLUGIN: ContentCategory.PLUGIN,
            ContentType.STRATEGY_PLUGIN: ContentCategory.PLUGIN,
            ContentType.TERMINAL_PLUGIN: ContentCategory.PLUGIN,
            ContentType.TEST_PLUGIN: ContentCategory.PLUGIN,
            ContentType.VARS_PLUGIN: ContentCategory.PLUGIN,
            ContentType.EDA_EVENT_SOURCE: ContentCategory.EXTENSION,
            ContentType.EDA_EVENT_FILTER: ContentCategory.EXTENSION,
        }.get(self)



import logging
import os

import hashlib

from galaxy_importer import exceptions as exc

log = logging.getLogger(__name__)


def sha256sum_from_fo(fo):
    block_size = 65536
    sha256 = hashlib.sha256()
    for block in iter(lambda: fo.read(block_size), b""):
        sha256.update(block)
    return sha256.hexdigest()


def sha256sum_from_path(filename):
    with open(filename, "rb") as fo:
        return sha256sum_from_fo(fo)


def check_artifact_file(path_prefix, artifact_file):
    """Check existences of artifact_file on fs and check the chksum matches

    Args:
        path_prefix (str): Any file path prefix we need to add to file paths in the
            CollectionArtifactFile artifact_file
        artifact_file (CollectionArtifactFile): object with the expected info about
            the file on the fs that will be checked.
            This info includes name, type, path, and checksum.

    Raises:
        CollectionArtifactFileNotFound: If artifact_file is not found on the file system.
        CollectionArtifactFileChecksumError: If the sha256sum of the on disk
            artifact_file contents does not match artifact_file.chksum_sha256.

    Returns:
        bool: True if artifact_file check is ok, otherwise should raise exception
    """
    log.debug("artifact_file: %s", artifact_file)

    artifact_file_path = os.path.join(path_prefix, artifact_file.name)
    if not os.path.exists(artifact_file_path):
        msg = f"The file ({artifact_file.name}) was not found"
        raise exc.CollectionArtifactFileNotFound(missing_file=artifact_file.name, msg=msg)

    actual_chksum = sha256sum_from_path(artifact_file_path)

    if actual_chksum != artifact_file.chksum_sha256:
        err_msg = (
            f"File {artifact_file.name} sha256sum should be "
            f"{artifact_file.chksum_sha256} but the actual sha256sum was {actual_chksum}"
        )
        log.error(err_msg)
        raise exc.CollectionArtifactFileChecksumError(err_msg)

    return True



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.


import logging
import os
import pkg_resources
import requests
import time
import uuid
import yaml

from galaxy_importer import config
from galaxy_importer import exceptions
from galaxy_importer.ansible_test.runners.base import BaseTestRunner


default_logger = logging.getLogger(__name__)

cfg = config.Config()
# TODO(cutwater): Implement individual timeouts for each step.
API_CHECK_RETRIES = int(os.environ.get("IMPORTER_JOB_API_CHECK_RETRIES", "300"))
API_CHECK_DELAY_SECONDS = int(os.environ.get("IMPORTER_JOB_API_CHECK_DELAY_SECONDS", "3"))
OCP_SERVICEACCOUNT_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/"
IMAGE_BASE_NAME = "ansible-test"


class OpenshiftJobTestRunner(BaseTestRunner):
    """Run image as an openshift job."""

    def run(self):
        self.log.info("Preparing job pod to run ansible-test sanity")

        job = Job(
            ocp_domain=os.environ["IMPORTER_API_DOMAIN"],
            namespace=os.environ["IMPORTER_JOB_NAMESPACE"],
            session_token=OpenshiftJobTestRunner._get_token(),
            ca_path=OpenshiftJobTestRunner._get_ca_path(),
            archive_url=self.file_url,
            job_template=OpenshiftJobTestRunner._get_job_template(),
            logger=self.log,
        )
        job.create()
        job.wait_on_pod_ready()
        iter_logs = job.get_logs()

        for line in iter_logs:
            if isinstance(line, bytes):
                self.log.error("Unexpected bytes in logs: {}".format(str(line)))
                continue
            self.log.info(line)

        job.cleanup()

    @staticmethod
    def _get_token():
        with open(os.path.join(OCP_SERVICEACCOUNT_PATH, "token"), "r") as f:
            token = f.read().rstrip()
        return token

    @staticmethod
    def _get_ca_path():
        return os.path.join(OCP_SERVICEACCOUNT_PATH, "ca.crt")

    @staticmethod
    def _get_job_template():
        path = pkg_resources.resource_filename("galaxy_importer", "ansible_test/job_template.yaml")
        with open(path, "r") as f:
            job_template = f.read()
        return job_template


class Job(object):
    """Interact with Openshift Job via REST API."""

    def __init__(
        self,
        ocp_domain,
        namespace,
        session_token,
        ca_path,
        job_template,
        archive_url,
        logger,
    ):
        self.name = "ansible-test-" + str(uuid.uuid4())
        self.auth_header = {"Authorization": f"Bearer {session_token}"}
        self.ca_path = ca_path
        self.jobs_url = f"{ocp_domain}/apis/batch/v1/namespaces/{namespace}/jobs"
        self.job_name_url = f"{self.jobs_url}/{self.name}"
        self.pods_url = f"{ocp_domain}/api/v1/namespaces/{namespace}/pods"
        self.pod_url_template = "{pods_url}/{pod_name}"
        self.log_url_template = "{pods_url}/{pod_name}/log"
        self.job_yaml = job_template.format(
            job_name=self.name,
            archive_url=archive_url,
            memory_request=os.environ.get("IMPORTER_MEMORY_REQUEST", "256Mi"),
            memory_limit=os.environ.get("IMPORTER_MEMORY_LIMIT", "1Gi"),
            cpu_request=os.environ.get("IMPORTER_CPU_REQUEST", "500m"),
            cpu_limit=os.environ.get("IMPORTER_CPU_LIMIT", "500m"),
            job_timeout=os.environ.get("IMPORTER_JOB_TIMEOUT", "900"),
            image_tag=os.environ.get("IMPORTER_IMAGE_TAG", "0b948de"),
        )
        self.log = logger or default_logger

    def create(self):
        """Create the job."""

        self.log.info(f"Creating job {self.name}")
        r = requests.post(
            self.jobs_url,
            headers=self.auth_header,
            json=yaml.safe_load(self.job_yaml),
            verify=self.ca_path,
        )
        if r.status_code != requests.codes.created:
            raise exceptions.AnsibleTestError(
                f"Could not create job: {r.status_code} {r.reason} {r.text}"
            )

    def wait_on_pod_ready(self):
        """Wait until job's pod initializes, pulls image, and starts running."""

        self.log.info("Creating pod...")
        for i in range(API_CHECK_RETRIES):
            pods = self.get_pods()
            if len(pods) < 1:
                time.sleep(API_CHECK_DELAY_SECONDS)
                continue
            break

        if len(pods) < 1:
            self.cleanup()
            raise exceptions.AnsibleTestError("Could not create pod associated with job")

        self.log.info("Scheduling pod and waiting until it is running...")
        for i in range(API_CHECK_RETRIES):
            pods = self.get_pods()
            pod_phase = pods[0]["status"]["phase"]
            if pod_phase != "Pending":
                return
            time.sleep(API_CHECK_DELAY_SECONDS)

        self.log.debug(pods[0]["status"])
        self.cleanup()
        raise exceptions.AnsibleTestError("Could not start pod associated with job")

    def get_pods(self):
        """Get pods associated with job."""
        params = {"labelSelector": f"job-name={self.name}"}
        r = requests.get(
            self.pods_url, headers=self.auth_header, params=params, verify=self.ca_path
        )
        try:
            pods = r.json()["items"]
        except (KeyError, ValueError):
            raise exceptions.AnsibleTestError("Could not access pod assocated with job")
        return pods

    @staticmethod
    def get_pod_name(pod):
        return pod["metadata"]["name"]

    def get_logs(self):
        """Returns stream of lines from the logs of the pod."""
        pod = self.get_pods()[0]
        pod_name = self.get_pod_name(pod)
        r = requests.get(
            url=f"{self.pods_url}/{pod_name}/log",
            headers=self.auth_header,
            params=dict(follow="true"),
            verify=self.ca_path,
            stream=True,
        )
        return r.iter_lines(decode_unicode=True)

    def cleanup(self):
        """Deletes job and any pods associated to it."""
        requests.delete(self.job_name_url, headers=self.auth_header, verify=self.ca_path)
        self.log.debug(f"Deleted job {self.name}")

        for pod in self.get_pods():
            pod_name = self.get_pod_name(pod)
            requests.delete(
                f"{self.pods_url}/{pod_name}",
                headers=self.auth_header,
                verify=self.ca_path,
            )
            self.log.debug(f"Deleted pod {pod_name}")

            status = pod["status"]["phase"]
            if status == "Failed":
                reason = pod["status"]["containerStatuses"][0]["state"]["terminated"]["reason"]
                raise exceptions.AnsibleTestError(
                    f'Pod terminated with status: "{status}" and reason: "{reason}"'
                )



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import argparse
import json
import logging
import os
import re
import sys

from galaxy_importer import collection, legacy_role
from galaxy_importer import config
from galaxy_importer.exceptions import ImporterError

FILENAME_REGEXP = re.compile(
    r"^(?P<namespace>\w+)-(?P<name>\w+)-" r"(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$"
)
logger = logging.getLogger(__name__)


def main(args=None):
    config_data = config.ConfigFile.load()
    cfg = config.Config(config_data=config_data)
    setup_logger(cfg)
    args = parse_args(args)

    data = call_importer(args, cfg=cfg)
    if not data:
        return 1

    if args.print_result:
        print(json.dumps(data, indent=4))

    write_output_file(data)


def setup_logger(cfg):
    """Sets up logger with custom formatter."""
    logger.setLevel(getattr(logging, cfg.log_level_main, "INFO"))

    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)


class CustomFormatter(logging.Formatter):
    """Formatter that does not display INFO loglevel."""

    def formatMessage(self, record):
        if record.levelno == logging.INFO:
            return "%(message)s" % vars(record)
        else:
            return "%(levelname)s: %(message)s" % vars(record)


def parse_args(args):
    parser = argparse.ArgumentParser(
        description="Run importer on collection and save result to disk."
    )
    parser.add_argument("file", nargs="?", help="artifact to import")
    parser.add_argument(
        "--git-clone-path",
        dest="git_clone_path",
        help="git directory with collection that will get built",
    )
    parser.add_argument(
        "--output-path",
        dest="output_path",
        help="path where built collection will be stored",
    )
    parser.add_argument(
        "--print-result",
        dest="print_result",
        action="store_true",
        help="print importer result to console",
    )
    parser.add_argument(
        "--legacy-role",
        dest="legacy_role",
        action="store_true",
        help="import a legacy role rather than collection",
    )
    parser.add_argument(
        "--namespace", dest="namespace", help="namespace of the legacy role to import"
    )
    return parser.parse_args(args=args)


def call_importer(args, cfg):  # pragma: no cover
    """Returns result of galaxy_importer import process.

    :param file: Artifact file to import.

    Method excluded from pytest unit test coverage, tests exist in tests/integration
    """
    if args.legacy_role:
        if args.file is None:
            logger.error("Must supply the directory of the role")
            return None
        if args.namespace is None:
            logger.error("Importing legacy role requires explicit namespace")
            return None
        try:
            data = legacy_role.import_legacy_role(args.file, args.namespace, logger=logger, cfg=cfg)
        except ImporterError as e:
            logger.error(f"The import failed for the following reason: {str(e)}")
            return None
        except Exception as e:
            logger.exception(f"Unexpected error occurred: {str(e)}")
            return None
    else:
        if not args.file:
            return collection.import_collection(
                git_clone_path=os.path.abspath(args.git_clone_path),
                output_path=os.path.abspath(args.output_path),
                logger=logger,
                cfg=cfg,
            )

        match = FILENAME_REGEXP.match(os.path.basename(args.file))
        namespace, name, version = match.groups()
        filename = collection.CollectionFilename(namespace, name, version)

        with open(args.file, "rb") as fh:
            try:
                data = collection.import_collection(fh, filename, logger=logger, cfg=cfg)
            except ImporterError as e:
                logger.error(f"The import failed for the following reason: {str(e)}")
                return None
            except Exception as e:
                logger.exception(f"Unexpected error occurred: {str(e)}")
                return None

    logger.info("Importer processing completed successfully")
    return data


def write_output_file(data):
    with open("importer_result.json", "w") as output_file:
        output_file.write(json.dumps(data, indent=4))


if __name__ == "__main__":
    exit(main())



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging
import os
import pkg_resources
import shutil
import tempfile

from galaxy_importer import exceptions
from shutil import copy
from subprocess import Popen, PIPE, STDOUT, run, CalledProcessError

default_logger = logging.getLogger(__name__)


class Build(object):
    """Use docker/podman to build ansible-test image with artifact inside."""

    def __init__(self, filepath, collection_name, cfg, logger=default_logger):
        self.cfg = cfg
        self.container_engine = Build.get_container_engine(cfg)
        self.filepath = filepath
        self.log = logger or default_logger
        self.image = ""
        self.working_dir = tempfile.TemporaryDirectory()

    def build_image(self):
        if self.container_engine == "docker":
            self.log.info("Building Dockerfile")
        else:
            self.log.info("Building ContainerFile")
        Build._build_dockerfile(self.working_dir.name)
        Build._copy_collection_file(dir=self.working_dir.name, filepath=self.filepath)

        self.log.info("Building image...")
        self.image = Build._build_image_with_artifact(
            container_engine=self.container_engine, dir=self.working_dir.name
        )
        return self.image

    def cleanup(self):
        self.log.info("Removing temporary files, image and container")
        self.working_dir.cleanup()

        cmd = [self.container_engine, "image", "rm", "-f", self.image]
        try:
            run(cmd)
        except CalledProcessError as e:
            raise exceptions.AnsibleTestError(
                "An exception occurred in: {}, message={}".format(" ".join(cmd), e.msg)
            )

    @staticmethod
    def get_container_engine(cfg):
        if cfg.local_image_docker is True:
            return "docker"
        else:
            return "podman"

    def _build_dockerfile(dir):
        pkg_dockerfile = pkg_resources.resource_filename(
            "galaxy_importer", "ansible_test/container/Dockerfile"
        )
        file_location = os.path.join(dir, "Dockerfile")
        shutil.copyfile(pkg_dockerfile, file_location)

        with open(file_location, "r+") as f:
            lines = f.readlines()
            for index, line in enumerate(lines):
                if "ENV HOME" in line:
                    # TODO move RUN command to base image
                    lines.insert(index - 1, "\nRUN chown -R user1:root /archive\n")
                    lines.insert(index - 1, "\nCOPY archive.tar.gz /archive/archive.tar.gz\n")
                    break
            f.seek(0)
            f.writelines(lines)

    def _build_image_with_artifact(container_engine, dir):
        pkg_entrypoint = pkg_resources.resource_filename(
            "galaxy_importer", "ansible_test/container/entrypoint.sh"
        )
        eda_linting = pkg_resources.resource_filename(
            "galaxy_importer", "ansible_test/container/eda/tox.ini"
        )
        shutil.copyfile(pkg_entrypoint, os.path.join(dir, "entrypoint.sh"))
        os.mkdir(os.path.join(dir, "eda"))
        shutil.copyfile(eda_linting, os.path.join(dir, "eda", "tox.ini"))

        cmd = [container_engine, "build", ".", "--quiet"]
        proc = Popen(
            cmd,
            cwd=dir,
            stdout=PIPE,
            stderr=STDOUT,
            encoding="utf-8",
        )

        image_id = ""
        for line in proc.stdout:
            image_id = line.strip()

        return_code = proc.wait()
        if return_code != 0:
            raise exceptions.AnsibleTestError(
                "An exception occurred in {}, returncode={}".format(" ".join(cmd), return_code)
            )
        if container_engine == "docker":
            image_id = image_id.split(":")[-1]
        return image_id

    def _copy_collection_file(dir, filepath):
        path = os.path.join(dir, "archive.tar.gz")
        copy(filepath, path)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import collections
import hashlib
import mimetypes
import os

import markdown
import bleach
from bleach_allowlist import markdown_tags, markdown_attrs

README_NAME = "README"
DOCFILE_EXTENSIONS = [
    ".md",
]
DOCFILE_MIMETYPES = {
    ".md": "text/markdown",
}
DOCFILE_MAX_SIZE = 512 * 1024  # 512 KiB

DocFile = collections.namedtuple("DocFile", ["name", "text", "mimetype", "hash"])

for _ext, _type in DOCFILE_MIMETYPES.items():
    mimetypes.add_type(_type, _ext)


class FileSizeError(Exception):
    pass


def get_readme_doc_file(directory):
    """Find and get readme file from directory.

    :return DocFile: Documentation file contents
    """
    filename = _find_readme(directory)
    if not filename:
        return None
    return _get_file(directory, filename)


def get_doc_files(directory):
    """Find and get list of documentation files from directory.

    :return: List of DocFile documentation files
    """
    filenames = _find_doc_files(directory)
    if not filenames:
        return None
    return [_get_file(directory, f) for f in filenames]


def get_html(doc_file):
    """Get html for a documentation file.

    :param doc_file: DocFile
    """
    if doc_file.mimetype == "text/markdown":
        return _render_from_markdown(doc_file)
    return None


def _find_readme(directory):
    """Look for and return valid readme file found in directory.

    :return str: Filename of readme
    """
    for ext in DOCFILE_EXTENSIONS:
        filename = os.path.join(directory, README_NAME + ext)
        if os.path.exists(filename):
            return filename
    return None


def _find_doc_files(directory):
    """Look for and return valid documentation files found in directory.

    :return str: List of filenames
    """
    result = []
    if not os.path.exists(directory):
        return result
    for filename in os.listdir(directory):
        if not os.path.isfile(os.path.join(directory, filename)):
            continue
        mimetype, _ = mimetypes.guess_type(filename)
        if mimetype not in DOCFILE_MIMETYPES.values():
            continue
        result.append(os.path.join(directory, filename))
    return result


def _get_file(directory, filename):
    """Get documentation file from directory.

    :return DocFile: Documentation file contents
    """
    if not os.path.exists(filename):
        return None

    if os.path.getsize(filename) > DOCFILE_MAX_SIZE:
        raise FileSizeError(
            'Documentation file "{0}" is bigger than {1} KiB.'.format(
                os.path.relpath(filename, directory),
                int(DOCFILE_MAX_SIZE / 1024),
            )
        )

    mimetype, encoding = mimetypes.guess_type(filename)

    with open(filename, "rb") as fp:
        raw_text = fp.read()
    hash_ = hashlib.sha256(raw_text).hexdigest()

    return DocFile(
        name=os.path.basename(filename),
        text=raw_text.decode("utf-8"),
        mimetype=mimetype,
        hash=hash_,
    )


def _render_from_markdown(doc_file):
    """Render html from markdown documentation file.

    :param doc_file: DocFile"""
    # notes on bleach coming after markdown, and bleach_allowlist pkg:
    # https://github.com/Python-Markdown/markdown/issues/225
    unsafe_html = markdown.markdown(doc_file.text, extensions=["extra"])
    return bleach.clean(
        unsafe_html,
        tags=markdown_tags + ["pre", "table", "thead", "th", "tr", "td"],
        attributes=markdown_attrs,
        styles=[],
        strip=True,
    )



import attr
import logging
import os

from galaxy_importer import config
from galaxy_importer import exceptions as exc
from galaxy_importer.loaders import LegacyRoleLoader
from galaxy_importer import __version__

default_logger = logging.getLogger(__name__)


def import_legacy_role(dirname, namespace, cfg=None, logger=None):
    """Process import on legacy role directory.

    NOTE: Must be run inside the parent directory of the role (parent directory of dirname).

    :param dirname: directory of role.
    :param namespace: namespace of role.
    :param cfg: Optional config.
    :param logger: Optional logger instance.

    :raises exc.ImporterError: On errors that fail the import process.

    :return: metadata of legacy role
    """

    logger = logger or default_logger
    logger.info(f"Importing with galaxy-importer {__version__}")

    # Ensure that dirname exists.
    if not os.path.exists(dirname):
        raise exc.ImporterError(f"The path '{dirname}' does not exist")

    # Ensure the input dirname is a subdirectory of the current directory.
    if dirname == os.curdir:
        raise exc.ImporterError(
            "Cannot run importer from role directory. Switch to parent directory"
        )
    joined = os.path.join(os.curdir, os.path.basename(os.path.normpath(dirname)))
    if not os.path.exists(joined):
        raise exc.ImporterError("Must run importer from parent directory of legacy role")
    if not os.path.isdir(joined):
        raise exc.ImporterError("Legacy role must be a directory")

    # Load user-specified or default configuration.
    if cfg is None:
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)

    return _import_legacy_role(dirname, namespace, cfg, logger)


def _import_legacy_role(dirname, namespace, cfg, logger):
    """Returns legacy role metadata."""

    data = LegacyRoleLoader(dirname, namespace, cfg, logger).load()
    logger.info("Legacy role loading complete")
    return attr.asdict(data)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.


class ImporterError(Exception):
    """Base class for importer exceptions"""

    pass


class AnsibleTestError(ImporterError):
    """Exception when running ansible-test."""

    pass


class ManifestNotFound(ImporterError):
    pass


class ManifestValidationError(ImporterError):
    pass


class ManifestFileListValidationError(ImporterError):
    pass


class FileNotInFileManifestError(ImporterError):
    """Files found in the artifact archive which were not listed in the file manifest"""

    def __init__(self, unexpected_files=None, msg=None):
        msg = msg or "Unexpected files found in the artifact but not the file manifest"
        super().__init__(msg)
        self.unexpected_files = unexpected_files or []


class CollectionArtifactFileNotFound(ImporterError):
    """The file the CollectionArtifactFile represents was not found"""

    def __init__(self, missing_file=None, msg=None):
        msg = msg or "File was listed in the file manifest but it was not found"
        super().__init__(msg)
        self.missing_file = missing_file


class CollectionArtifactFileChecksumError(ImporterError):
    """The chksum of the file contents does not match chksum_sha256sum"""

    pass


class ContentFindError(ImporterError):
    pass


class ContentLoadError(ImporterError):
    pass


class ContentNameError(ImporterError):
    pass


class FileParserError(ImporterError):
    pass


class LegacyRoleSchemaError(ImporterError):
    pass



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import collections
import itertools
import logging
import os

import attr

from galaxy_importer import constants
from galaxy_importer.file_parser import ExtensionsFileParser


default_logger = logging.getLogger(__name__)

Result = collections.namedtuple("Result", ["content_type", "path"])

ROLE_SUBDIRS = ["tasks", "vars", "handlers", "meta"]


class ContentFinder(object):
    """Searches for content in directories inside collection."""

    def find_contents(self, path, logger=None):
        """Finds contents in path and return the results.

        :rtype: Iterator[Result]
        :return: Iterator of find results.
        """

        self.path = path
        self.log = logger or default_logger

        self.log.info("Finding content inside collection")
        contents = self._find_content()

        try:
            first = next(contents)
        except StopIteration:
            return []
        else:
            return itertools.chain([first], contents)

    def _find_content(self):
        for content_type, directory, func in self._content_type_dirs():
            content_path = os.path.join(self.path, directory)
            if not os.path.exists(content_path):
                continue
            yield from func(content_type, content_path)

    def _find_plugins(self, content_type, content_dir):
        """Find all python files anywhere inside content_dir."""
        for path, _, files in os.walk(content_dir):
            for file in files:
                if not file.endswith(".py") or file == "__init__.py":
                    continue
                file_path = os.path.join(path, file)
                rel_path = os.path.relpath(file_path, self.path)
                yield Result(content_type, rel_path)

    def _find_roles(self, content_type, content_dir):
        """Find all dirs inside roles dir where contents match a role."""

        def is_dir_a_role(current_dir):
            """Check for contents indicating directory is a role."""
            _, dirs, _ = next(os.walk(current_dir))
            if set(ROLE_SUBDIRS) & set(dirs):
                return True
            return False

        def recurse_role_dir(path):
            """Iterate over all subdirs and yield roles."""
            if is_dir_a_role(path):
                rel_path = os.path.relpath(path, self.path)
                yield Result(content_type, rel_path)
                return
            path, dirs, _ = next(os.walk(path))
            for dir in dirs:
                yield from recurse_role_dir(os.path.join(path, dir))

        yield from recurse_role_dir(content_dir)

    def _content_type_dirs(self):
        for content_type in constants.ContentType:
            if content_type == constants.ContentType.ROLE:
                yield content_type, "roles", self._find_roles
            elif content_type == constants.ContentType.MODULE:
                yield content_type, "plugins/modules", self._find_plugins
            else:
                yield (
                    content_type,
                    "plugins/" + content_type.value,
                    self._find_plugins,
                )

        for content_type, full_path in self._get_ext_types_and_path():
            yield content_type, full_path, self._find_plugins

    def _get_ext_types_and_path(self):
        extension_dirs = ExtensionsFileParser(self.path).get_extension_dirs()

        if not extension_dirs:
            return []

        # remove ext_dir not currently allowed in the content list
        for ext_dir in list(extension_dirs):
            if ext_dir not in constants.ALLOWED_EXTENSION_DIRS:
                self.log.warning(
                    f"The extension type '{ext_dir}' listed in 'meta/extensions.yml' is "
                    "custom and will not be listed in Galaxy's contents nor documentation"
                )
                extension_dirs.remove(ext_dir)

        content_types_and_dirs = [
            (constants.ContentType(dir), f"extensions/{dir}") for dir in extension_dirs
        ]

        return content_types_and_dirs


@attr.s
class FileWalker(object):
    collection_path = attr.ib()
    file_errors = attr.ib(factory=list)

    def walk(self):
        full_collection_path = os.path.abspath(self.collection_path)

        for dirpath, dirnames, filenames in os.walk(
            full_collection_path, onerror=self.on_walk_error, followlinks=False
        ):
            for dirname in dirnames:
                dir_full_path = os.path.join(dirpath, dirname)
                yield dir_full_path

            for filename in filenames:
                full_path = os.path.join(dirpath, filename)
                yield full_path

    def on_walk_error(self, walk_error):
        default_logger.warning("walk error on %s: %s", walk_error.filename, walk_error)
        self.file_errors.append(walk_error)



import logging
import os
import shutil
from subprocess import PIPE, Popen, TimeoutExpired

from galaxy_importer import config
from galaxy_importer import constants
from galaxy_importer import exceptions as exc
from galaxy_importer.utils import markup
from galaxy_importer import schema

default_logger = logging.getLogger(__name__)


class LegacyRoleLoader(object):
    """Loads legacy role information."""

    def __init__(self, dirname, namespace, cfg=None, logger=None):
        self.dirname = dirname
        self.namespace = namespace

        # If no config is found, use default configuration values.
        self.cfg = cfg
        if self.cfg is None:
            config_data = config.ConfigFile.load()
            self.cfg = config.Config(config_data=config_data)

        self.log = logger or default_logger

        self.name = None
        self.metadata = None
        self.content_obj = None
        self.readme = None
        self.readme_html = None

    def load(self):
        """Loads role metadata and content information and lints role."""

        self._validate_namespace()

        self.metadata = self._load_metadata()
        self.name = self._load_name()
        self.readme = self._load_readme()
        self.readme_html = markup.get_html(self.readme)

        if self.cfg.run_ansible_lint:
            self._lint_role()

        return schema.LegacyImportResult(
            namespace=self.namespace,
            name=self.name,
            metadata=self.metadata,
            readme_file=self.readme.name,
            readme_html=self.readme_html,
        )

    def _validate_namespace(self):
        """Validate the namespace is a valid github username."""

        if constants.GITHUB_USERNAME_REGEXP.match(self.namespace) is None:
            raise exc.ImporterError(f"namespace {self.namespace} is invalid")

    def _load_metadata(self):
        """Loads role metadata."""

        # Search for metadata in paths
        # meta.yml, meta.yaml, meta/main.yml, and meta/main.yaml.
        meta_path = None
        for file in constants.ROLE_META_FILES:
            path = os.path.join(self.dirname, file)
            if os.path.exists(path):
                meta_path = path
                break
        if meta_path is None:
            raise exc.ImporterError("Metadata not found at any path")

        return schema.LegacyMetadata.parse(meta_path)

    def _load_name(self):
        """Determine the name of a legacy role."""

        # The name of the role is determined by the name of its directory
        # UNLESS there is a galaxy_info.role_name field. The metadata field
        # overrides the directory name.
        if self.metadata.galaxy_info.role_name is not None:
            name = self.metadata.galaxy_info.role_name
        else:
            name = os.path.basename(os.path.normpath(self.dirname))

        self.log.info(f"Determined role name to be {name}")

        return name

    def _load_readme(self):
        """Find the README file of a role."""

        readme = markup.get_readme_doc_file(self.dirname)
        if not readme:
            raise exc.ImporterError("No role readme found")

        return readme

    def _lint_role(self):
        """Log ansible-lint output.

        ansible-lint stdout are linter violations, they are logged as warnings.
        ansible-lint stderr includes info about vars, file discovery,
        summary of linter violations, config suggestions, and raised errors.
        Only raised errors are logged, they are logged as errors.
        """

        self.log.info(f"Linting role {self.name} via ansible-lint...")

        if not shutil.which("ansible-lint"):
            self.log.warning("ansible-lint not found, skipping lint of role")
            return

        cmd = [
            "/usr/bin/env",
            f"ANSIBLE_LOCAL_TEMP={self.cfg.ansible_local_tmp}",
            "ansible-lint",
            "--profile",
            "production",
            "--parseable",
            "--nocolor",
            self.dirname,
        ]
        if self.cfg.offline_ansible_lint:
            cmd.append("--offline")

        self.log.debug("CMD:", "".join(cmd))

        proc = Popen(cmd, encoding="utf-8", stdout=PIPE, stderr=PIPE)

        try:
            outs, errs = proc.communicate(timeout=120)
        except TimeoutExpired:
            self.log.error("Timeout on call to ansible-lint")
            proc.kill()
            outs, errs = proc.communicate()

        for line in outs.splitlines():
            self.log.warning(line.strip())

        for line in errs.splitlines():
            if line.startswith(constants.ANSIBLE_LINT_ERROR_PREFIXES):
                self.log.error(line.strip())

        self.log.info("...ansible-lint run complete")
